"""
Persist VLBA metadata extracted from .idifits files
"""

import os
import sys
import datetime
import re
import logging
import hashlib
import time

from pycapo import CapoConfig
from .VLBAFITSUtil import VLBAFits
from .schema import create_session
from . import proposalutils, archiveutils
from . import NGASUtil as ng
from .schema.model import Project, Filegroup, File, ExecutionBlock, Scan, Subscan, Author, Intent, Configuration, DataDescription, Polarization, ScienceProduct

session = create_session('VLBA')
LOG = logging.getLogger('ingestion')
config = CapoConfig()

def persistVLBAMetadata(fits_file):
    LOG.info(f"Persisting fits file: {fits_file}")
    # Basically want to bail out of the whole process if the file name is already in the Files table
    starttime = time.time()
    vlba_fits = VLBAFits(fits_file)
    LOG.info(f"Fits file completed in {time.time() - starttime}s")
    starttime = time.time()
    projectcode = vlba_fits.getproject()
    segment = vlba_fits.getsegment()
    correlator_pass = vlba_fits.getcorrelatorpass()
    # vlba_fits.getscantable().pprint(max_lines=250, max_width=250)

    project = persistproject(projectcode)
    # Persist archive location information
    correlations_filegroup, correlator_pass_filegroup = persistfilegroups(project, segment, correlator_pass)

    file = persistfile(correlator_pass_filegroup, fits_file)

    # Now the science data
    executionblock = persistexecblock("VLBA", project, segment, correlations_filegroup, vlba_fits, file)
    if executionblock is not None:
        if project.starttime is None or project.starttime > executionblock.starttime:
            project.starttime = executionblock.starttime
        if project.endtime is None or project.endtime < executionblock.endtime:
            project.endtime = executionblock.endtime

    session.add(project)

    session.commit()

    LOG.info(f"Database updates for file {fits_file} completed in {time.time() - starttime}s")


def persistproject(projectcode):

    existingproject = archiveutils.getproject(projectcode)
    if existingproject is None:
        searchproposalcode = re.sub('([A-Z]+)0', '\g<1>', projectcode)
        pstproposal = proposalutils.getproposal(searchproposalcode)
        pstauthors = persistauthors(projectcode)
        if pstproposal is None:
            psttitle = ""
            pstabstract = ""
            if projectcode.startswith('DQ'):
                # Data analyst test project
                proprietaryduration = 0
            else:
                proprietaryduration = 365
        else:
            if pstproposal.PROPOSAL_TYPE == "Director's Discretionary Time":
                proprietaryduration = round(365 / 2, 0)
            else:
                proprietaryduration = 365
            psttitle = pstproposal.TITLE
            pstabstract = pstproposal.ABSTRACT

        project = Project(project_code           = projectcode,
                      opt_project_id         = 0,
                      legacy_id              = projectcode,
                      title                  = psttitle,
                      abstract               = pstabstract,
                      total_observation_time = None,
                      last_addition          = datetime.datetime.utcnow(),
                      starttime              = None,
                      endtime                = None,
                      proprietary_duration   = proprietaryduration)
        project.authors = pstauthors
        session.add(project)
        return project
    return existingproject


def persistauthors(project):
    searchproposalcode = re.sub('([A-Z]+)0', '\g<1>', project)
    pstauthors = proposalutils.getauthors(searchproposalcode)
    projauthors = []
    if pstauthors is None or len(pstauthors) == 0:
        author = Author(project_code=project,
                        username='operations',
                        firstname='VLBA',
                        lastname='Operations',
                        pst_person_id=10154,
                        is_pi=True)
        session.add(author)
        projauthors.append(author)
    else:
        for pstauthor in pstauthors:
            projauthors.append(Author(project_code=project, username=pstauthor.user.personName, firstname=pstauthor.FIRST_NAME, lastname=pstauthor.LAST_NAME,
                                  pst_person_id=pstauthor.person_id, is_pi=proposalutils.ispi(pstauthor)))

    return projauthors


def persistfilegroups(project, segment, correlator_pass):
    correlations_filegroup = archiveutils.get_vlba_correlations_filegroup(project.project_code, segment, 'correlations')

    if correlations_filegroup is not None:
        correlator_pass_filegroup = archiveutils.get_vlba_correlator_pass_filegroup(project.project_code, segment, 'correlator_pass')
        session.add(correlator_pass_filegroup)
        correlations_filegroup = correlator_pass_filegroup.parent_filegroup
        if correlator_pass_filegroup is not None:
            return correlations_filegroup, correlator_pass_filegroup
        else:
            filegroup = Filegroup(parent_filegroup = correlations_filegroup,
                                datasize = 0,
                                type='correlator_pass'
                                  )
            session.add(filegroup)
            return correlations_filegroup, filegroup
    else:
        parent_filegroup = Filegroup(project=project,
                              datasize = 0,
                              type = 'correlations'
                              )
        filegroup = Filegroup(parent_filegroup = parent_filegroup,
                              datasize = 0,
                              type='correlator_pass'
                              )
        session.add(parent_filegroup, filegroup)
        return parent_filegroup, filegroup



def persistfile(filegroup, fits_file):
    file_path = os.path.basename(fits_file)
    ngas_id = os.path.basename(fits_file)
    existingfile = archiveutils.getvlbafile(ngas_id)
    site_name = config.getstring('archive-ingestion.NGASSite')
    cluster_name = config.getstring('archive-ingestion.NGASCluster')
    if existingfile is None:
        file = File(file_path=file_path,
                    ngas_id=ngas_id,
                    ngas_location=site_name,
                    ngas_cluster=cluster_name,
                    filegroup1=filegroup,
                    filename=os.path.basename(fits_file),
                    filesize=os.path.getsize(fits_file),
                    format=fits_file.suffix[1:].upper(),
                    type=fits_file.suffix[1:].upper(),
                    checksum=sha256(fits_file),
                    checksum_type='SHA256',
                    ingestion_time=datetime.datetime.utcnow())

        file.filegroup1.datasize = file.filegroup1.datasize + file.filesize
        session.add(file)
        return file
    LOG.critical(f"File {ngas_id} already exists. Exiting.")
    exit(-1)


def persistexecblock(telescope, project, segment, filegroup, vlba_fits, file):
    execblock = archiveutils.getvlbaexecutionblock(project.project_code, segment)
    science_product_locator = ng.generate_science_product_locator(telescope, 'execblock')
    if execblock is None:
        execblock = ExecutionBlock(
           ost_exec_block_id=0,
           filegroup=filegroup,
           calibration_level=0,
           telescope='VLBA',
           configuration=None,
           scheduling_block_id=0,
           project=project,
           segment=segment,
           ngas_fileset_id=None,
           starttime=0,
           endtime=0,
           calibration_status='Do Not Calibrate',
           scheduling_block_type=None,
           band_code=vlba_fits.lookup_observing_band(),
           science_product_locator=science_product_locator)

        science_product = ScienceProduct(
            science_product_locator=science_product_locator,
            science_product_type="Execution Block",
            metadata_ingestion_date=datetime.datetime.utcnow(),
            metadata_ingestion_version=1,
            filegroup=filegroup,
            external_name=project.project_code+segment,
            external_system="VLBA Operations",
            execution_block=execblock
        )

        execblock.project.science_products.append(science_product)

        session.add(execblock)
        session.add(science_product)

        starttime, endtime = persistscans(vlba_fits, execblock, filegroup, file)
        execblock.starttime = starttime
        execblock.endtime = endtime
    else:
        starttime, endtime = persistscans(vlba_fits, execblock, filegroup, file)
        if starttime < execblock.starttime:
            execblock.starttime = starttime
        if endtime > execblock.endtime:
            execblock.endtime = endtime

    return execblock



def persistscans(vlba_fits, execution_block, filegroup, file):
    minstarttime, maxendtime = 1000000000.0, 0.0
    scans = vlba_fits.getscantable()
    for scan_number in range(len(scans)):
        scan = Scan(
            execution_block=execution_block,
            filegroup=filegroup,
            min_bandwidth = 0.0,
            max_bandwidth = 0.0,
            min_frequency = 0.0,
            max_frequency = 0.0,
            polarization_code = 0
        )
        minfrequency, maxfrequency, bandwidth, starttime, endtime, polarization = persistsubscans(execution_block, scan, scans[scan_number], file)

        # MHz --> Hz
        scan.min_frequency = minfrequency * 1000000
        scan.max_frequency = maxfrequency * 1000000

        scan.min_bandwidth = bandwidth
        scan.max_bandwidth = bandwidth
        # scan.polarization_code = polarization.polarization_id
        session.add(scan)
        if starttime < minstarttime:
            minstarttime = starttime
        if endtime > maxendtime:
            maxendtime = endtime
    return minstarttime, maxendtime

def persistsubscans(execution_block, scan, scandata, file):
    configuration = persistconfiguration(execution_block, 0)
    subscan = Subscan(
        scan=scan,
        obstype='',
        starttime=scandata['START_TIME'],
        endtime=scandata['END_TIME'],
        sourcename=scandata['SOURCE'].split('\x00', 1)[0], # Fix some source names with terminating nulls
        sourcetype='STAR',
        ra=scandata['RA'],
        dec=scandata['DEC'],
        exposure_time=scandata['TIME_ON_SOURCE'],
        integration_time=scandata['INTEGRATION_TIME'],
        receiver_id=0,
        backend='DIFX',
        configuration=configuration,
        intent='TRACK',
        file=file
    )

    bandwidth = scandata['BANDWIDTH']
    frequencies = [float(freq) for freq in scandata['FREQUENCY'].split(',')]
    minfrequency = min(frequencies)
    maxfrequency = max(frequencies)
    polarization = scandata['POLARIZATION']
    for frequency in frequencies:
        persistdatadescription(subscan, configuration, frequency, bandwidth, polarization)

    session.add(subscan)
    return minfrequency, maxfrequency, bandwidth, subscan.starttime, subscan.endtime, polarization


def persistdatadescription(subscan, configuration, frequency, bandwidth, polarization):
    local_polarization = session.merge(polarization)
    datadescription = DataDescription(
        bandwidth=bandwidth,
        frequency=frequency,
        polarization=local_polarization,
        configuration=configuration
    )
    session.add(datadescription)


def persistconfiguration(execution_block, configuration):
    configuration = Configuration(
        execution_block=execution_block,
        configuration=configuration
    )
    session.add(configuration)
    return configuration


def sha256(fname):
    hash_sha256 = hashlib.sha256()
    with open(fname, "rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            hash_sha256.update(chunk)
    return hash_sha256.hexdigest()


if __name__ == '__main__':
    path = sys.argv[1]
    fits_file = sys.argv[2]
    persistVLBAMetadata(os.path.join(path, fits_file))
