from typing import Dict, List, Union

import json

from workspaces.workflow.enum import WorkflowRequestState
from workspaces.workflow.schema import (
    Workflow,
    WorkflowRequest,
    WorkflowRequestFile,
    WorkflowTemplate,
)
from workspaces.workflow.schema_interfaces import WorkflowIF
from workspaces.workflow.services.interfaces import WorkflowInfoIF


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def lookup_workflow_request(self, request_id: int) -> WorkflowRequest:
        return self.session.query(WorkflowRequest).filter_by(workflow_request_id=request_id).first()

    def lookup_workflow_templates_for(self, workflow_name: str) -> List[WorkflowTemplate]:
        """
        Queries the workflow_templates table for all templates belonging to a given workflow

        :param workflow_name: Name of workflow to lookup templates for
        :return: List of templates
        """
        return self.session.query(WorkflowTemplate).filter_by(workflow_name=workflow_name).all()

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()

    def all_workflow_requests(self) -> List[WorkflowRequest]:
        """
        Query for the list of all workflow requests

        :return: List of workflow requests
        """
        return self.session.query(WorkflowRequest).all()

    def create_workflow_request(
        self, workflow: Union[str, WorkflowIF], argument: Dict
    ) -> WorkflowRequest:
        """
        Create new workflow request and save to database

        :param workflow: workflow to run
        :param argument: workflow arguments
        :return: new WorkflowRequest
        """

        # if it's an object with a workflow_name, then that's what we use,
        # otherwise, the argument must be already be a string
        workflow_name = workflow.workflow_name if hasattr(workflow, "workflow_name") else workflow

        if any(item in workflow_name for item in ["null", "download"]):
            argument["need_project_metadata"] = False
        else:
            argument["need_project_metadata"] = True

        request = WorkflowRequest(
            workflow_name=workflow_name,
            argument=argument,
            state=WorkflowRequestState.Created.name,
            results_dir="",
        )
        self.save_request(request)
        print(f"Created Workflow Request {request}")
        return request

    def save_request(self, request: WorkflowRequest):
        """
        Save a given request and return an integer identifier for it

        :param request: request to save
        """
        self.session.add(request)
        self.session.flush()

    def save_file(
        self, request: WorkflowRequest, filename: str, content: bytes
    ) -> WorkflowRequestFile:
        """
        Save a given file for the specified request to the database

        :param request: request to update with new file
        :param filename: filename of new file
        :param content: contents of new file in bytes
        :return:
        """
        wrf = WorkflowRequestFile(
            workflow_request_id=request.workflow_request_id,
            filename=filename,
            content=content,
        )
        self.session.add(wrf)
        self.session.flush()
        return wrf
